<?php if (!defined('PmWiki')) exit();
/*  Copyright 2004 Patrick R. Michaud (pmichaud@pobox.com)
    This file is part of PmWiki; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  See pmwiki.php for full details.


    Modifications Copyright 2005-2010 John Rankin and Donald Gordon
    This is a complete rewrite of the markup rules to produce wikibook xml
*/

## first we preserve text in [=...=] and [@...@]
function LatexPreserveText($sigil, $text, $lead) {
  if ($sigil=='=') return $lead.Keep(str_replace(array("`","'"),array('&grave;','&sq;'),$text));
  if (strpos($text, "\n")===false) 
    return "$lead<tbook:verb>".Keep($text)."</tbook:verb>";
#    return "$lead<tbook:visual markup='tt'>".Keep($text)."</tbook:visual>";
  $text = preg_replace("/^[^\\S\n]*\n/", "\n", substr($lead,1).$text);
  $text = preg_replace("/\n[^\\S\n]*$/", "\n", $text);
  return "$lead<tbook:verbatim>".Keep($text)."</tbook:verbatim>";
}

Markup('[=','_begin',"/(\n[^\\S\n]*)?\\[([=@])(.*?)\\2\\]/se",
    "LatexPreserveText('$2', PSS('$3'), '$1')");

## \\$  (end of line joins)
Markup('\\$','>nl1',"/\\\\(?>(\\\\*))\n/e",
       "(strlen('$1') > 1 ? '<tbook:newline vspace=\"0.5em\" />' :".
       " (strlen('$1') > 0 ? '<tbook:newline />' : ''))");

## (:linebreaks:)
Markup('linebreaks', 'directives',
  '/\\(:(no)?linebreaks:\\)/ei',
  "PZZ(\$GLOBALS['HTMLPNewline'] = ('$1'!='no') ? '<tbook:newline  />' : '')");

## (:keywords:), then (:description:) have to come before meta tags
Markup('keywords', '>&', "/\\(:keywords?\\s+([^'\n]+?):\\)/ei",
  "PZZ(array_unshift(\$GLOBALS['HTMLHeaderFmt'],PSS('<keywords>$1</keywords>')))");
Markup('description', '>&', "/\\(:description\\s+(.+?):\\)/ei",
  "PZZ(InsertAbstractTag(PSS('<abstract><p>$1</p></abstract>')))");

function InsertAbstractTag($abstract) {
  global $HTMLHeaderFmt;
  if (strstr($HTMLHeaderFmt[0], '<keywords>'))
      array_splice($HTMLHeaderFmt,1,0,$abstract);
  else
      array_unshift($HTMLHeaderFmt,$abstract);
}

#### inline markups ####
## ''emphasis''
Markup("''",'inline',"/''(.*?)''/",'<tbook:em>$1</tbook:em>');

## '''strong'''
Markup("'''","<''","/'''(.*?)'''/","<tbook:visual markup='bf'>\$1</tbook:visual>");

## '''''strong emphasis'''''
Markup("'''''","<'''","/'''''(.*?)'''''/",'<tbook:visual markup="bf"><tbook:em>$1</tbook:em></tbook:visual>');

## @@code@@
Markup('@@','inline','/@@(.*?)@@/','<tbook:visual markup="tt">$1</tbook:visual>');

## '+big+', '-small-'
Markup("'+","<'''''","/'\\+(.*?)\\+'/",
  '<tbook:visual markup="big">$1</tbook:visual>');
Markup("'-","<'''''","/'\\-(.*?)\\-'/",
  '<tbook:visual markup="small">$1</tbook:visual>');

## '^superscript^', '_subscript_'
Markup("'^","<'''''","/'\\^(.*?)\\^'/",'<tbook:visual markup="sup">$1</tbook:visual>');
Markup("'_","<'''''","/'_(.*?)_'/",'<tbook:visual markup="sub">$1</tbook:visual>');

## [+big+], [-small-]

function LimitPDFFontSize($n) {
  if ($n < -4) $n = -4;
  if ($n > 5) $n = 5;
  return $n;
}

Markup('[+','inline','/\\[(([-+])+)(.*?)\\1\\]/e',
  "'<tbook:style size=\"'.LimitPDFFontSize($2strlen('$1')).'\">'.
    PSS('$3').'</tbook:style>'");

## {+ins+}, {-del-}
SDV($WPInsertFmt, '$Text'); SDV($WPDeleteFmt, '');
Markup('{+','inline','/\\{\\+(.*?)\\+\\}/e',
  "str_replace('\$Text',PSS('$1'),\$GLOBALS['WPInsertFmt'])");
Markup('{-','inline','/\\{-(.*?)-\\}/e',
  "str_replace('\$Text',PSS('$1'),\$GLOBALS['WPDeleteFmt'])");

## [[<<]] (break)
Markup('[[<<]]','inline','/\\[\\[&lt;&lt;\\]\\]/','<tbook:newline />');

#### Links ####
## images as targets for links (hi-res substitution)
SDV($SubstituteHiResImg['arrow'], true);
SDV($SubstituteHiResImg['pipe'], false);
if ($SubstituteHiResImg['arrow']) { 
  Markup('hires1','>inline',
  "/\[\[(.*?)(\"[^\"]*\")?\s*-&gt;\s*([^\s$UrlExcludeChars]+$ImgExtPattern)\]\]/e", 
  "SubstituteHiResImg(PSS('$3'),PSS('$1'),PSS('$2'))");
  $PdfImgTagFmt = "<tbook:graphics kind='vector' file='\$LinkUrl' />";
  Markup('hires3','>inline',
  "/\[\[(\\L)([^\s$UrlExcludeChars]+)$ImgExtPattern(\"[^\"]*\")?\s*-&gt;\s*\\1\\2\.pdf\]\]/e", 
  "Keep(\$GLOBALS['LinkFunctions']['$1'](\$pagename,'$1','$2.pdf',NULL,'$1$2.pdf',
      \$GLOBALS['PdfImgTagFmt']), 'L') . ('$3' ? PSS(' tbook:caption=$3 ') : '')");
} else Markup('hires1','>inline',
  "/\[\[(.*?)(\"[^\"]*\")?\s*-&gt;\s*([^\s$UrlExcludeChars]+$ImgExtPattern)\]\]/e", 
  "PreserveLoResImg(PSS('$3'),PSS('$1'),PSS('$2'))");
if ($SubstituteHiResImg['pipe']) Markup('hires2','>inline',
  "/\[\[([^\s$UrlExcludeChars]+$ImgExtPattern)\s*\|\s*(.*?)(\"[^\"]*\")?\]\]/e",
  "SubstituteHiResImg(PSS('$1'),PSS('$2'),PSS('$3'))");
else Markup('hires2','>hires1',
  "/\[\[([^\s$UrlExcludeChars]+$ImgExtPattern)\s*\|\s*(.*?)(\"[^\"]*\")?\]\]/e",
  "PreserveLoResImg(PSS('$1'),PSS('$2'),PSS('$3'))");
function SubstituteHiResImg($img, $txt, $alt) {
  global $UrlExcludeChars, $ImgExtpattern;
  return (preg_match("/^[^\s$UrlExcludeChars]+$ImgExtPattern$/", $txt)) ?
       "$img$alt" : "$img $txt";
}
function PreserveLoResImg($img, $txt, $alt) {
  global $UrlExcludeChars, $ImgExtpattern;
  return (preg_match("/^[^\s$UrlExcludeChars]+$ImgExtPattern$/", $txt)) ?
       "$txt$alt" : "[[$txt -&gt; $img]]";
}

## [[#anchor]]
Markup('[[#','<[[','/(?>\\[\\[#([A-Za-z][-.:\\w]*))\\]\\]/e',
  "TrackAnchors(\"\$pagename.$1\") ? '' :
  Keep(\"<tbook:wrap id='\$pagename.$1'></tbook:wrap>\",'L')");
Markup('[[#wp','<[[#','/(?>\\[\\[#(wikipublisher\.[-.:\\w]*))\\]\\]/e',
  "Keep(\"<tbook:wrap id='$1'></tbook:wrap>\",'L')");

## inline images with captions
Markup('caption|', '<[[|',"/(\[\[[^|\]]*\|\s*)(.*?$ImgExtPattern)\"([^\"]+)\"(\]\])/",
  '$2 | $1$3$4');
Markup('caption->','<[[|',"/(\[\[)([^\]]*$ImgExtPattern)\"([^\"]+)\"(\s*-&gt;\s*.*?\]\])/",
  '$2 | $1$3$4');
Markup('caption1','<img',"/($ImgExtPattern)(\"[^\"]+\")/",
  '$1 tbook:caption=$2 ');
$ImgTagFmt = '<tbook:graphics kind="bitmap" file="$LinkUrl" />';

## Lines that begin with displayed images receive their own block. A pipe
## following the image indicates a "caption" (generates a caption tag).
Markup('^img', '<block',
  "/^((?>(\\s+|%%|%[A-Za-z][-,.=:#\\w\\s'\"]*%)*))($KeepToken(\\d+L)$KeepToken)(.*)$/e",
  "(strpos(\$GLOBALS['KPV']['$4'],'<tbook:graphics')===false) ? PSS('$0') :
  ((strpos(\$GLOBALS['KPV']['$4'],'\'page\'') || strpos(\$GLOBALS['KPV']['$4'],'\"page\"')) ?
  PSS('$0') : '<:block>'.SetFigureCaption(PSS('$1'),PSS('$3'),PSS('$5')).'</tbook:figure>')");

function SetFigureCaption($style,$img,$captext) {
  global $KeepToken, $KPV;
  preg_match('/^(?:\\s*tbook:caption="([^"]+)")?\\s*(\\|)?\\s*(.*?)(%%)?$/',
      $captext, $match);
  list($capstr, $alt, $cap, $text, $sty) = $match;
  
  $captag = 'tbook:caption'; $figtag = 'tbook:figure'; $subtag = 'tbook:subfigure';

# Check for an image block (subfigures)
  $ImageAndCaptionPat = 
    "((?>(\\s+|%%|%[A-Za-z][-,.=:#\\w\\s'\"]*%)*))($KeepToken(\\d+L)$KeepToken)".
    '(?:\\s*tbook:(caption="[^"]+"\\s*))?';
  $trial = $text; $icount = 0; $dim = 'width="$Width" >';
  $subfig = "<$subtag ".'caption="'.$alt.'" '.$dim.$img."</$subtag>";
  while (preg_match("/^$ImageAndCaptionPat/",$trial,$mat) && strpos($KPV[$mat[4]],'tbook:graphics')) {
    $icount++;
    $subfig  .= $mat[1]."<$subtag ".$mat[5].$dim.$mat[3]."</$subtag>";
    $trial = str_replace($mat[0],'',$trial);
  }
  $width  = (1-0.014*$icount)/(1+$icount);
  if ($icount) return
        "<$figtag>$style".str_replace('$Width',"$width",$subfig)."<$captag>$trial$sty</$captag>";

# There is just a regular captioned image
  if ($cap) {
      $caption = ($text) ? $text : (($alt)  ? $alt : '-');
      $lof     = ($alt)  ? $alt  : '';
  } else {
      $caption = ($alt)  ? $alt  : (($text) ? '-'  : '' );
      $lof     = $caption;
      if ($text) $caption .= "<tbook:newline vspace='0.5em' />".
              preg_replace("/^<tbook:newline.*?>/",'',$text);
  }
  return "$style<$figtag>$img".(($caption) ? "<$captag".(($lof) ? ' caption="'.$lof.'" ' : ''). 
                 ">$caption$sty</$captag>" : $sty);
}

# clean up caption text
Markup('caption2','>^img','/\s*tbook:caption="([^"]+)"\s*/',' ');

#### Block markups ####
# unblocked lines w/block markup become anonymous <:block>
Markup('^!<:', '<^<:',
  '/^(?!<:)(?=.*<\\/?tbook:(table|p|itemize|enumerate|description|h|blockquote|verbatim)\\b)/',
  '<:block>');

# plus a special for his <div>s
Markup('^<div>', '<^<:','/^(?!<:)(?=.*<div\\b)/','<:block>');

$BlockMarkups['p'] = array('<tbook:p>','','</tbook:p>',0);
$BlockMarkups['ul'] =
  array('<tbook:itemize><tbook:item>','</tbook:item><tbook:item>','</tbook:item></tbook:itemize>',1);
$BlockMarkups['ol'] =
 array('<tbook:enumerate><tbook:item>','</tbook:item><tbook:item>','</tbook:item></tbook:enumerate>',1);
$BlockMarkups['indent'] = array('<tbook:blockquote><tbook:p>','</tbook:p><tbook:p>','</tbook:p></tbook:blockquote>',1);
$BlockMarkups['outdent'] = array('<tbook:p style=\'hanging\'>','</tbook:p><tbook:p style=\'hanging\'>','</tbook:p>',1);
$BlockMarkups['pre'] = array('<tbook:verbatim>','','</tbook:verbatim>',0);
$BlockMarkups['description'] = array('<tbook:description>','</tbook:item>','</tbook:item></tbook:description>',1);

## definition lists
Markup('^::','block','/^(:+)([^:<]*(<[^>]+>[^:<]*)*):/','<:description,$1><tbook:term>$2</tbook:term><tbook:item>');

## honour list spacing (to enable, remove $HTMLVSpace = ''; from template php)
$HTMLVSpace = '<vspace/>';
$Transition['vspace'] = 0;
Markup('skipii','>^<:',
    "/<vspace\/>(<\/tbook:item><tbook:(?:term|item)[^>]*)>/",'$1 skip="small">');
Markup('skippi','>^<:',
  "/<vspace\/>((?:<!--.*-->)?<tbook:(?:itemize|enumerate|description)[^>]*>)/",
    '<tbook:p skip="med" />$1');
Markup('skipip','>^<:',
    "/(<\/tbook:(?:itemize|enumerate|description|blockquote)>)<vspace\/>(<tbook:p\b)(\s+skip=)?/e",
    "'$1$2'.('$3' ? '$3' : ' skip=\"med\"')");
## trick to fool advanced table patch
Markup('skipp','>^<:','/^(<tbook:p)>/','$1 >');

## Q: and A:
Markup('^Q:', 'block', '/^Q:\s*(.*)$/',
#    "<:block><tbook:p skip='big'><tbook:visual markup='bf'>".
#    "<tbook:style size='1'>$1</tbook:style></tbook:visual></tbook:p>");
    "<:block><tbook:h level='4'>$1</tbook:h>");
Markup('^A:', 'block', '/^A:\s*(.*)$/', "<:block><tbook:p>$1</tbook:p>");

## tables
## ||cell||, ||!header cell||, ||!caption!||
Markup('^||||','block',
  '/^\\|\\|.*\\|\\|.*$/e',
  "FormatTbookTableRow(PSS('$0'))");

# ||table attributes
Markup('^||','>^||||','/^\\|\\|(.*)$/e',
  "PZZ(\$GLOBALS['BlockMarkups']['table'][0] = '<tbook:tabular '. PQA(TbookTableAttrs(PSS('$1'))).'><tbook:tabbody>')");

function TbookTableAttrs($x) {
  if (preg_match('/id=([^\s]*)/',$x,$m)) $r  = 'id="'.$m[1].'" '; else $r = '';
  if (preg_match('/class=.?(long|wide|big)/',$x,$c))   $r .= 'class="'.$c[1].'"';
  if (preg_match('|(\ssummary="[^"]*")|',$x,$m))   $r .= $m[1];
  return $r;
}
$BlockMarkups['table'] = array("<tbook:tabular><tbook:tabbody>",'','</tbook:tabbody></tbook:tabular>',0);

function FormatTbookTableRow($x) {
  global $Block, $TableCellAttrFmt, $MarkupFrame, $TableRowAttrFmt,
    $TableRowIndexMax, $FmtV;
  static $rowcount,$RowspanCount,$ColspanCount;
  $x = preg_replace('/\\|\\|$/','',$x);
  $td = explode('||',$x); $y=''; $rsc=0;
  for($i=0;$i<count($td);$i++) {
    $tdpre=''; $tdpost='';
    if ($td[$i]=='') continue;
    $FmtV['$TableCellCount'] = $i;
    $attr = FmtPageName($TableCellAttrFmt, '');
    $td[$i] = preg_replace('/^(!?)\\s+$/', '$1&nbsp;', $td[$i]);
    if (preg_match('/^!(.*?)!$/',$td[$i],$match))
      { $td[$i]=$match[1]; $t='caption'; $attr='';  }
    elseif (preg_match('/^!(.*)$/',$td[$i],$match))
      { $td[$i]=$match[1]; $t='cell';
        if (preg_match('/^\s*.{0,16}\+*\s*$/', $td[$i])) $attr .= " wrap='no'";
        $attr .= " title='true'";
        $tdpre = "<tbook:visual markup='bf'>";
        $tdpost = "</tbook:visual>"; }
    else {
      if (preg_match('/^\s*.{0,16}\+*\s*$/', preg_replace('/<.*?>/','',$td[$i]))) 
          $attr .= " wrap='no'";
      $t='cell';
    }
    if ($t != 'caption') {
      if (@$RowspanCount[$i]>1) {
          for ($j=0;$j<$ColspanCount[$i];$j++) {
              $y .= "<tbook:$t wrap='no'></tbook:$t>"; $rsc++;
          }
          $RowspanCount[$i]--; 
      }
      if (preg_match('/^\\s.*\\s$/',$td[$i])) { $attr .= " align='center'"; }
      elseif (preg_match('/^\\s/',$td[$i])) { $attr .= " align='right'"; }
      elseif (preg_match('/\\s$/',$td[$i])) { $attr .= " align='left'"; }
      for ($colspan=1;$i+$colspan<count($td);$colspan++)
        if ($td[$colspan+$i]!='') break;
      if ($colspan>1) { $attr .= " colspan='$colspan'"; }
      if (preg_match('/^(.*?)(\+\++)(\\s*)$/',$td[$i],$match)) {
          $td[$i]=$match[1].$match[3]; $attr .= " rowspan='".strlen($match[2])."'";
          $RowspanCount[$i+$rsc] = strlen($match[2]);
          $ColspanCount[$i+$rsc] = $colspan;
      }
    }
    $y .= "<tbook:$t $attr>".$tdpre.$td[$i].$tdpost."</tbook:$t>";

  }
  if ($t=='caption') return "<:table,1>$y";
  if (@$RowspanCount[$i+$rsc]>1) {
      for ($j=0;$j<$ColspanCount[$i+$rsc];$j++) $y .= "<tbook:$t wrap='no'></tbook:$t>"; 
      $RowspanCount[$i+$rsc]--; 
  }
  if ($MarkupFrame[0]['cs'][0] != 'table') { $rowcount = 0; $RowspanCount = array(); }
  else $rowcount++;
  $FmtV['$TableRowCount'] = $rowcount + 1;
  $FmtV['$TableRowIndex'] = ($rowcount % $TableRowIndexMax) + 1;
  $trattr = FmtPageName($TableRowAttrFmt, $pagename);
  return "<:table,1><tbook:row $trattr>$y</tbook:row>";
}

## nested heading formulation (excluding subparagraph)
function BlockIsOpen() {
  global $MarkupFrame;
  $close = $MarkupFrame[0]['closeall'];
  foreach (array_keys((array)$close) as $k)
    if (strncmp($k, 'div', 3) == 0) return true;
    elseif ($k == 'table') return true;
  return false;
}

function HeadSection($level, $sec='section') {
  global $wikibook_section,$wikibook_subsection,$wikibook_subsubsec,
         $section_open,$subsec_open,$subsub_open,$adjust_heading_levels,
         $MarkupFrame, $TbookStructure;
  SDV($wikibook_section, 0);
  $t = 'tbook:'; $s = $q = '';
  $subsec = $TbookStructure[$sec]; $subsub = $TbookStructure[$subsec];
  if ($wikibook_section==0) {
    $wikibook_section = $level;
    $section_open = false;
    $wikibook_subsection = ($sec=='subparagraph') ? 9 : 0;
  }
  if ($level==$wikibook_section) {
    $s = "<$t$sec>";
    if ($section_open)
        $s = ($subsec_open ? ($subsub_open ? "</$t$subsub>" : '') . 
            "</$t$subsec>" : '') . "</$t$sec>$s";
    else $section_open = true;
    $subsec_open = false;
  } elseif ($level>$wikibook_section || $adjust_heading_levels) {
        if ($wikibook_subsection==0) {
            $wikibook_subsection = $level;
            $wikibook_subsubsec  = ($subsec=='subparagraph') ? 9 : 0;
        }
        if ($level==$wikibook_subsection) {
            $s = "<$t$subsec>";
            if ($subsec_open)
                $s = ($subsub_open ? "</$t$subsub>" : '') . "</$t$subsec>$s";
            else $subsec_open = true;
            $subsub_open = false;
        } elseif ($level>$wikibook_subsection || $adjust_heading_levels) {
            if ($wikibook_subsubsec==0) {
                $wikibook_subsubsec = $level;
            }
            if ($level==$wikibook_subsubsec) {
                if (!$subsec_open) {
                    $c = 'heading';
                    $s = "<$t$subsec><$t$c>%red%?%%</$t$c>";
                    $subsec_open = true;
                    $subsub_open = false;
                }
                $s .= "<$t$subsub>";
                if ($subsub_open) $s = "</$t$subsub>$s";
                else $subsub_open = true;
            }
        }
  } else $q = ' %red%?%%';
  $MarkupFrame[0]['closeall']['section'] = 
     ($subsec_open ? ($subsub_open ? "</$t$subsub>" : '') . 
        "</$t$subsec>" : '') . "</$t$sec>(:closesection:)";
  $r = ($level==$wikibook_section || $level==$wikibook_subsection || 
        $level==$wikibook_subsubsec) ?
        ($c = 'heading') : ($c = 'h') . " level='$level'";
  return array("$s<$t$r>", "$q</$t$c>");
}

function HeadStructure($pagename,$level) {
  global $HeadingDepth;
  return (BlockIsOpen()) ? array("<tbook:h level='$level'>", "</tbook:h>") : 
    HeadSection($level, $HeadingDepth[$pagename]);
}
function EscapeUS($text) {
  return str_replace('_','&us;',$text);
}

$TbookStructure = array(
    'section'       => 'subsection',
    'subsection'    => 'subsubsection',
    'subsubsection' => 'paragraph',
    'paragraph'     => 'subparagraph',
    'subparagraph'  => '*',
    '*'             => '**');

Markup('^!','block','/^(!{1,6})\\s?(.*)$/e',
  "'<:block>'.implode(EscapeUS(PSS('$2')),HeadStructure(\$pagename,strlen('$1')))");
Markup('closesection', '>closeall', 
  '/^(<:block>)(.*?)\\(:closesection:\\)(.*?)$/', '$1$3$2');
$MarkupFrameBase['posteval']['section'] = "\$GLOBALS['wikibook_section'] = 0;";

## horiz rule (when style is ding, wikipublisher uses a character value)
SDV($WPhrStyle,  'ding');
if ($WPhrStyle) {
    Markup('^----','>^->','/^----+/','<:block><tbook:hr style="'.$WPhrStyle.'" />');
} else
    Markup('^----','>^->','/^----+/','<:block><tbook:hr />');

#### (:table:) markup (AdvancedTables)

## patch: correct (:cell:) or (:div:)followed by unblocked text
Markup('cellpatch','<table',"/^(\(:cell(?:nr)?(?:\s.*?)?:\))(.+?)$/",
    '$1<tbook:p>$2 </tbook:p>');
Markup('divpatch','>links',
    "/^(\(:div(?:\s.*?)?:\)(?:%.*?apply=div%)?)(.+?)$/",
    '$1<tbook:p>$2</tbook:p>');
Markup('headnr','>cellpatch', "/^\(:head((?:nr)?(?:\s.*?)?):\)(.*?)$/e",
    "'(:cell'.PSS('$1').' style=head:)'.('$2' ? '<tbook:p>'.PSS('$2').' </tbook:p>' : '')");

if (@$Transition['nodivnest']) {
function TbookCells($name,$attr) {
  global $MarkupFrame;
  $attr = preg_replace('/([a-zA-Z]=)([^\'"]\\S*)/',"\$1'\$2'",$attr);
  $tattr = @$MarkupFrame[0]['tattr'];
  $name = strtolower($name);
  $out = array('<:block>');
  if (strncmp($name, 'cell', 4) != 0 || @$MarkupFrame[0]['closeall']['div']) {
    $out[] = @$MarkupFrame[0]['closeall']['div']; 
    unset($MarkupFrame[0]['closeall']['div']);
    $out[] = @$MarkupFrame[0]['closeall']['table']; 
    unset($MarkupFrame[0]['closeall']['table']);
  }
  if ($name == 'div') {
    $MarkupFrame[0]['closeall']['div'] = "</tbook:group>";
    $out[] = "<tbook:group $attr>";
  }
  if ($name == 'table') $MarkupFrame[0]['tattr'] = $attr .' tabletype="advanced"';
  if (strncmp($name, 'cell', 4) == 0) {
    if (!@$MarkupFrame[0]['closeall']['table']) {
       $MarkupFrame[0]['closeall']['table'] = "</tbook:cell></tbook:row></tbook:tabbody></tbook:advtabular>";
       $out[] = "<tbook:advtabular $tattr><tbook:tabbody><tbook:row><tbook:cell $attr>";
    } else if ($name == 'cellnr') $out[] = "</tbook:cell></tbook:row><tbook:hline /><tbook:row><tbook:cell $attr>";
    else $out[] = "</tbook:cell><tbook:cell $attr>";
  }
  return implode('', $out);
}

Markup('table', '<block',
  '/^\\(:(table|cell|cellnr|tableend|div|divend)(\\s.*?)?:\\)/ie',
  "TbookCells('$1',PSS('$2'))");
} else {
function TbookCellsNested($name,$attr) {
  global $MarkupFrame;
  static $RowspanCount, $RowspanAttr, $ColspanCount, $col, $rsc;
  $div = 'tbook:group'; $table = 'tbook:advtabular'; $tb = 'tbook:tabbody'; 
  $tr  = 'tbook:row';   $td    = 'tbook:cell';       $hl = 'tbook:hline';
  $rs  = '<tbook:p>&#8942;</tbook:p>';
  $attr = preg_replace('/([a-zA-Z]=)([^\'"]\\S*)/',"\$1'\$2'",$attr);
  $tattr = @$MarkupFrame[0]['tattr'];
  $name = strtolower($name);
  $key = preg_replace('/end$/', '', $name);
  if (strncmp($key, 'cell', 4) == 0) $key = 'cell';
  $out = '<:block>'.MarkupClose($key);
  if (substr($name, -3) == 'end') { 
      if (@$RowspanCount[$col+$rsc]>1) {
          $o = "</$td><$td ".$RowspanAttr[$col+$rsc].">$rs";
          $RowspanCount[$col+$rsc]--;
          $out = str_replace('<:block>','<:block>'.$o,$out);
      }
      return $out; 
  }
  $cf = & $MarkupFrame[0]['closeall'];
  if ($name == 'table') 
      $MarkupFrame[0]['tattr'] = $attr.' tabletype="advanced"'; 
  else if (strncmp($name, 'cell', 4) == 0) {
    if (!@$cf['table']) {
       $tattr = @$MarkupFrame[0]['tattr'];
       $out .="<$table $tattr><$tb><$tr>";
       $cf['table'] = "</$td></$tr></$tb></$table>";
       $col = 0; $rsc = 0; $RowspanCount = array();
    } else if ($name == 'cellnr') {
      if (@$RowspanCount[$col+$rsc]>1) {
          $out .= "</$td><$td ".$RowspanAttr[$col+$rsc].">$rs";
          $RowspanCount[$col+$rsc]--; 
      }
      $out .= "</$td></$tr><$hl /><$tr>";
      $col = 0; $rsc = 0;
    } else $out .= "</$td>";
    if (@$RowspanCount[$col]>1) {
        $out .= "<$td ".$RowspanAttr[$col].">$rs</$td>"; 
        $rsc += $ColspanCount[$col];
        $RowspanCount[$col]--; 
    }
    $colspan = 1;
    if (preg_match('/colspan=[\'"](\d+)[\'"]/',$attr,$mat)) $colspan = $mat[1];
    if (preg_match('/rowspan=[\'"](\d+)[\'"]/',$attr,$mat)) {
       $RowspanCount[$col+$rsc] = $mat[1];
       $ColspanCount[$col+$rsc] = $colspan;
       $attr = str_replace(' '.$mat[0], '', $attr);
       $RowspanAttr[$col+$rsc] = $attr;
    }
    $out .= "<$td $attr>";
    $cf['cell'] = ''; $col++;
  } else {
    if (preg_match('/div(\d+)/',$name,$match)) $attr .= " level='$match[1]'";
    $out .= "<$div $attr>";
    $cf[$key] = "</$div>";
  }
  return $out;
}


Markup('table', '<block',
  '/^\\(:(table|cell|cellnr|tableend|div\\d*(?:end)?)(\\s.*?)?:\\)/ie',
  "TbookCellsNested('$1',PSS('$2'))");
}
Markup('^>>', '<table',
  '/^&gt;&gt;(.*?)&lt;&lt;(.*)$/',
  '(:div:)%$1 apply=div%$2 ');

#### special stuff ####
## (:markup:) for displaying markup examples

function TbookMarkupMarkup($pagename, $lead, $text, $opt = '') {
  $MarkupMarkupOpt = array('class' => 'vert');
  $opt = array_merge($MarkupMarkupOpt, ParseArgs($opt));
  $class = preg_replace('/[^-\\s\\w]+/', ' ', @$opt['class']);
  $xml = MarkupToWikibook($pagename, $text, array('escape' => 0));
  $xml = str_replace('<tbook:table>','<tbook:table class="long">',$xml);
  $caption = (@$opt['caption']) ?  '<tbook:caption>'.$opt['caption'].'</tbook:caption>' : '';
  if (strpos($class, 'horiz') !== false) {
      $cap =  array('o' => '<tbook:group class="markup">'.$caption,
                    'c' => '</tbook:group>');
      return "$lead<:block>".Keep($cap['o']."<tbook:group class='cell' width='.5'>" .
             "<tbook:verbatim>" . wordwrap($text, 35) . "</tbook:verbatim>" .
        "\n</tbook:group><tbook:group class='cell' width='.5'>\n$xml\n</tbook:group>".$cap['c']);
  }
  return "$lead<:block>".Keep("<tbook:group class='frame'>" .
        "$caption<tbook:verbatim>" . 
        wordwrap($text, 66) . "</tbook:verbatim><tbook:hr style='dotted' />" .
        "\n$xml</tbook:group>");
}

Markup('markup', '<[=',
  "/(^|\\(:nl:\\))\\(:markup(\\s+([^\n]*?))?:\\)[^\\S\n]*\\[([=@])(.*?)\\4\\]/seim",
  "TbookMarkupMarkup(\$pagename, '$1', PSS('$5'), PSS('$3'))");
Markup('markupend', '>markup',
  "/(^|\\(:nl:\\))\\(:markup(\\s+([^\n]*?))?:\\)[^\\S\n]*\n(.*?)\\(:markupend:\\)/seim",
  "TbookMarkupMarkup(\$pagename, '$1', PSS('$4'), PSS('$3'))");

## fixups
function pNewlineSkip($pattr, $space) {
  $hrsp = '<tbook:hr style="space" />';
  return (preg_match('/skip=["\'].*?["\']/', $pattr, $a)) ?
      str_replace($a[0], 'skip="big"', $pattr) : ($space ? "$hrsp$pattr" : "$pattr skip='med'");
}
Markup('fixnewline1','>_end',
    "/(<tbook:p(?:\s+[^>]+)?)><tbook:newline\s*(?:(vspace=).*?)?\/>/e",
    "pNewLineSkip(PSS('$1'),'$2').'>'");
Markup('fixnewline2','>fixnewline1',
    '/<tbook:newline \/>\s*(<\/tbook:(blockquote|p))/','$1');
Markup('fixnewline3','>fixnewline2',
    '/^<tbook:newline \/>/','&nbsp; $0');

$MarkupFrameBase['posteval']['detag'] =
    "\$out=newStyles(\$out);".
    "\$out=preg_replace('/<(?!\/tbook:)(?!tbook:)(?!\!)[^>]+>/s','',\$out);".
    "\$out=preg_replace('/(<\/?)tbook:([^>]+>)/s','$1$2',\$out); ".

## fix appendix sections
    "\$out = preg_replace('/(<section)(><heading>[Aa]ppendix[:\s])/','$1 class=\"appendix\"$2',\$out); ".

## fix figure caption floats
    "\$out = preg_replace(\"/(<\/caption><\/figure>)\n?<p\s*>([^\n]*)\n?".
          "(<newline)\s*\/>\n?([^\n]*\n)?<\/p>/se\",".
          "\"'$3 vspace=\'0.5em\' />'.PSS('$2$1').('$4' ? '<p>$4</p>' : '')\",".
          "\$out); ".

## use summary if no caption
    "\$out = preg_replace('/(<(?:adv)?tabular[^>]*?)\s+summary='.\"['\\\"]([^\\\"]*?)['\\\"]\".'([^>]*><tabbody>)(<caption[^>]*)?/e',".
          "\"PSS('$1$3').('$4' ? '$4 caption=\\\"$2\\\"' : '<caption>$2</caption>')\",\$out);".

## move !table caption!
    "\$out = preg_replace('/(<(?:adv)?tabular[^>]*?class='.\"['\\\"](?:long|big)['\\\"]\".'[^>]*><tabbody>)(<caption[^>]*>.*?<\/caption>)".
                      "(.*?<\/(?:adv)?tabular>)/s','<table class=\"long\">$1$3$2</table>',\$out); ".
    "\$out = preg_replace('/(<(?:adv)?tabular[^>]*?class='.\"['\\\"]wide['\\\"]\".'[^>]*><tabbody>)(<caption[^>]*>.*?<\/caption>)".
                      "(.*?<\/(?:adv)?tabular>)/s','<table class=\"wide\">$1$3$2</table>',\$out); ".
    "\$out = preg_replace('/(<tabular[^>]*><tabbody>)(<caption[^>]*>.*?<\/caption>)".
                      "(.*?<\/tabular>)/s','<table>$1$3$2</table>',\$out); ".
    "\$out = preg_replace('/(<advtabular[^>]*><tabbody>)(<caption[^>]*>.*?<\/caption>)".
                      "(.*?<\/advtabular>)/s','<table>$1$3$2</table>',\$out); ".
    "\$out = preg_replace('/(<table[^>]*)(><(?:adv)?tabular\s+[^>]*)(id=[^>\s]*)/s', '$1 $3$2',\$out); ".
    "\$out = analyseTables(\$out);";

?>
